@@ -82,6 +82,11 @@ if (-f project_vars.sh) then
    source project_vars.sh
 endif
 
+# Get power and ground bus names
+if (-f ${synthdir}/${rootname}_powerground) then
+   source ${synthdir}/${rootname}_powerground
+endif
+
 # Prepend techdir to leffile unless leffile begins with "/"
 set abspath=`echo ${leffile} | cut -c1`
 if ( "${abspath}" == "/" ) then
@@ -106,15 +111,26 @@ else
    set spicepath=${techdir}/${spicefile}
 endif
 
-# logfile should exist, but just in case. . .
+if (!($?logdir)) then
+   set logdir=${projectpath}/log
+endif
+mkdir -p ${logdir}
+set lastlog=${logdir}/synth.log
+set synthlog=${logdir}/place.log
+rm -f ${synthlog} >& /dev/null
+rm -f ${logdir}/sta.log >& /dev/null
+rm -f ${logdir}/route.log >& /dev/null
+rm -f ${logdir}/post_sta.log >& /dev/null
 touch ${synthlog}
+set date=`date`
+echo "Qflow placement logfile created on $date" > ${synthlog}
 
 #----------------------------------------------------------
 # Done with initialization
 #----------------------------------------------------------
 
 # Check if last line of log file says "error condition"
-set errcond = `tail -1 ${synthlog} | grep "error condition" | wc -l`
+set errcond = `tail -1 ${lastlog} | grep "error condition" | wc -l`
 if ( ${errcond} == 1 ) then
    echo "Synthesis flow stopped on error condition.  Placement will not proceed"
    echo "until error condition is cleared."
@@ -127,16 +143,32 @@ endif
 
 cd ${projectpath}
 
-echo "Running blif2cel.tcl" |& tee -a ${synthlog}
-
-${scriptdir}/blif2cel.tcl --blif ${synthdir}/${rootname}.blif \
+echo "Running blif2cel to generate input files for graywolf" |& tee -a ${synthlog}
+if ( "$techleffile" == "" ) then
+    echo "blif2cel.tcl --blif ${synthdir}/${rootname}.blif --lef ${lefpath} --cel ${layoutdir}/${rootname}.cel" |& tee -a ${synthlog}
+    ${scriptdir}/blif2cel.tcl --blif ${synthdir}/${rootname}.blif \
 	--lef ${lefpath} --cel ${layoutdir}/${rootname}.cel >>& ${synthlog}
+else
+    echo "blif2cel.tcl --blif ${synthdir}/${rootname}.blif --lef ${techlefpath} --lef ${lefpath} --cel ${layoutdir}/${rootname}.cel" |& tee -a ${synthlog}
+    ${scriptdir}/blif2cel.tcl --blif ${synthdir}/${rootname}.blif \
+	--lef ${techlefpath} --lef ${lefpath} \
+	--cel ${layoutdir}/${rootname}.cel >>& ${synthlog}
+endif
+
+set errcond = $status
+if ( ${errcond} != 0 ) then
+   echo "blif2cel.tcl failed with exit status ${errcond}" |& tee -a ${synthlog}
+   echo "Premature exit." |& tee -a ${synthlog}
+   echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+   exit 1
+endif
 
 #---------------------------------------------------------------------
 # Spot check:  Did blif2cel produce file ${rootname}.cel?
 #---------------------------------------------------------------------
 
-if ( !( -f ${layoutdir}/${rootname}.cel || ( -M ${layoutdir}/${rootname}.cel \
+if ( !( -f ${layoutdir}/${rootname}.cel || \
+	( -f ${layoutdir}/${rootname}.cel && -M ${layoutdir}/${rootname}.cel \
 	< -M ${rootname}.blif ))) then
    echo "blif2cel failure:  No file ${rootname}.cel." |& tee -a ${synthlog}
    echo "blif2cel was called with arguments: ${synthdir}/${rootname}.blif "
@@ -155,9 +187,19 @@ endif
 cd ${layoutdir}
 
 if ( ${?initial_density} ) then
-   echo "Running decongest to set initial density of ${initial_density}"
+   echo "Running decongest to set initial density of ${initial_density}" \
+		|& tee -a ${synthlog}
+   echo "decongest.tcl ${rootname} ${lefpath} ${fillcell} ${initial_density}" \
+		|& tee -a ${synthlog}
    ${scriptdir}/decongest.tcl ${rootname} ${lefpath} \
 		${fillcell} ${initial_density} |& tee -a ${synthlog}
+   set errcond = $status
+   if ( ${errcond} != 0 ) then
+	 echo "decongest.tcl failed with exit status ${errcond}" |& tee -a ${synthlog}
+	 echo "Premature exit." |& tee -a ${synthlog}
+	 echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+	 exit 1
+   endif
    cp ${rootname}.cel ${rootname}.cel.bak
    mv ${rootname}.acel ${rootname}.cel
 endif
@@ -199,20 +241,6 @@ else
    echo "continuing without pin placement hints" |& tee -a ${synthlog}
 endif
 
-# Add fill cells for the power bus stripes
-# (this is work in progress, commented out for now)
-
-# echo "Running powerbus to add spacers for power bus stripes" |& tee -a ${synthlog}
-# ${scriptdir}/powerbus.tcl ${rootname} ${lefpath} ${fillcell} |& tee -a ${synthlog}
-
-# powerbus.tcl creates a .acel file if successful.  If not, then
-# leave the .cel file in place
-
-if ( -f ${rootname}.acel && ( -M ${rootname}.acel >= -M ${rootname}.cel )) then
-   cp ${rootname}.cel ${rootname}.cel.bak
-   mv ${rootname}.acel ${rootname}.cel
-endif
-
 #-----------------------------------------------
 # 1) Run GrayWolf
 #-----------------------------------------------
@@ -226,14 +254,23 @@ if ( !( ${?graywolf_options} )) then
 endif
 
 echo "Running GrayWolf placement" |& tee -a ${synthlog}
-   ${bindir}/graywolf ${graywolf_options} $rootname >>& ${synthlog}
+echo "graywolf ${graywolf_options} $rootname" |& tee -a ${synthlog}
+${bindir}/graywolf ${graywolf_options} $rootname >>& ${synthlog}
+
+set errcond = $status
+if ( ${errcond} != 0 ) then
+   echo "graywolf failed with exit status ${errcond}" |& tee -a ${synthlog}
+   echo "Premature exit." |& tee -a ${synthlog}
+   echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+   exit 1
 endif
 
 #---------------------------------------------------------------------
 # Spot check:  Did GrayWolf produce file ${rootname}.pin?
 #---------------------------------------------------------------------
 
-if ( !( -f ${rootname}.pin || ( -M ${rootname}.pin < -M ${rootname}.cel ))) then
+if ( !( -f ${rootname}.pin || \
+	( -f ${rootname}.pin && -M ${rootname}.pin < -M ${rootname}.cel ))) then
    echo "GrayWolf failure:  No file ${rootname}.pin." |& tee -a ${synthlog}
    echo "Premature exit." |& tee -a ${synthlog}
    echo "Synthesis flow stopped due to error condition." >> ${synthlog}
@@ -282,7 +319,8 @@ if ($makedef == 1) then
    # Spot check:  Did qrouter produce file ${rootname}.info?
    #---------------------------------------------------------------------
 
-   if ( !( -f ${rootname}.info || ( -M ${rootname}.info < -M ${rootname}.pin ))) then
+   if ( !( -f ${rootname}.info || \
+	( -f ${rootname}.info && -M ${rootname}.info < -M ${rootname}.pin ))) then
       echo "qrouter (-i) failure:  No file ${rootname}.info." |& tee -a ${synthlog}
       echo "Premature exit." |& tee -a ${synthlog}
       echo "Synthesis flow stopped due to error condition." >> ${synthlog}
@@ -295,7 +333,8 @@ if ($makedef == 1) then
    # been disabled, or else we'll try passing $fillcell directly to
    # place2def
 
-   echo "Running getfillcell.tcl" |& tee -a ${synthlog}
+   echo "Running getfillcell to determine cell to use for fill." |& tee -a ${synthlog}
+   echo "getfillcell.tcl $rootname ${lefpath} $fillcell" |& tee -a ${synthlog}
    set usefillcell = `${scriptdir}/getfillcell.tcl $rootname \
 	${lefpath} $fillcell | grep fill= | cut -d= -f2`
 
@@ -306,18 +345,37 @@ if ($makedef == 1) then
 
    # Run place2def to turn the GrayWolf output into a DEF file
 
+   if ( !( ${?place2def_options} )) then
+      set place2def_options = ""
+   endif
+
+   echo "Running place2def to translate graywolf output to DEF format." \
+		|& tee -a ${synthlog}
    if ( ${?route_layers} ) then
+      echo "place2def.tcl $rootname $usefillcell ${route_layers} ${place2def_options}" \
+		|& tee -a ${synthlog}
       ${scriptdir}/place2def.tcl $rootname $usefillcell ${route_layers} \
-		 >>& ${synthlog}
+		${place2def_options} >>& ${synthlog}
+      set errcond = $status
+      if ( ${errcond} != 0 ) then
+	 echo "place2def.tcl failed with exit status ${errcond}" |& tee -a ${synthlog}
+	 echo "Premature exit." |& tee -a ${synthlog}
+	 echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+	 exit 1
+      endif
    else
-      ${scriptdir}/place2def.tcl $rootname $usefillcell >>& ${synthlog}
+      echo "place2def.tcl $rootname $usefillcell ${place2def_options}" \
+		|& tee -a ${synthlog}
+      ${scriptdir}/place2def.tcl $rootname $usefillcell ${place2def_options} \
+		>>& ${synthlog}
    endif
 
    #---------------------------------------------------------------------
    # Spot check:  Did place2def produce file ${rootname}.def?
    #---------------------------------------------------------------------
 
-   if ( !( -f ${rootname}.def || ( -M ${rootname}.def < -M ${rootname}.pin ))) then
+   if ( !( -f ${rootname}.def || \
+	( -f ${rootname}.def && -M ${rootname}.def < -M ${rootname}.pin ))) then
       echo "place2def failure:  No file ${rootname}.def." |& tee -a ${synthlog}
       echo "Premature exit." |& tee -a ${synthlog}
       echo "Synthesis flow stopped due to error condition." >> ${synthlog}
@@ -328,16 +386,31 @@ if ($makedef == 1) then
    # Add spacer cells to create a straight border on the right side
    #---------------------------------------------------------------------
 
-   if ( -f ${scriptdir}/addspacers.tcl ) then
+   if ( !(${?nospacers}) && (-f ${scriptdir}/addspacers.tcl) ) then
 
       if ( !( ${?addspacers_options} )) then
          set addspacers_options = ""
       endif
 
-      echo "Running addspacers.tcl ${addspacers_options} ${rootname} ${lefpath} ${fillcell}" |& tee -a ${synthlog}
+      if ( "$techleffile" != "" ) then
+         set addspacers_options = "-techlef ${techlefpath} ${addspacers_options}"
+      endif
+
+      echo "Running addspacers to generate power stripes and align cell right edge" \
+		|& tee -a ${synthlog}
+      echo "addspacers.tcl ${addspacers_options} ${rootname} ${lefpath} ${fillcell}" \
+		|& tee -a ${synthlog}
 
       ${scriptdir}/addspacers.tcl ${addspacers_options} \
 		${rootname} ${lefpath} ${fillcell} >>& ${synthlog}
+      set errcond = $status
+      if ( ${errcond} != 0 ) then
+	 echo "addspacers.tcl failed with exit status ${errcond}" |& tee -a ${synthlog}
+	 echo "Premature exit." |& tee -a ${synthlog}
+	 echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+	 exit 1
+      endif
+
       if ( -f ${rootname}_filled.def ) then
 	 mv ${rootname}_filled.def ${rootname}.def
 	 # Copy the .def file to a backup called "unroute"
@@ -365,10 +438,11 @@ if ($makedef == 1) then
 
    # Create the main configuration file
 
-   # Variables "via_pattern" (none, normal, invert) and "via_stacks"
-   # can be specified in the tech script, and are appended to the
-   # qrouter configuration file.  via_stacks defaults to 2 if not
-   # specified.  It can be overridden from the user's .cfg2 file.
+   # Variables "via_pattern" (none, normal, invert), "via_stacks",
+   # and "via_use" can be specified in the tech script, and are
+   # appended to the qrouter configuration file.  via_stacks defaults
+   # to 2 if not specified.  It can be overridden from the user's .cfg2
+   # file.
 
    if (${scripting} == "T") then
       echo "# qrouter runtime script for project ${rootname}" > ${rootname}.cfg
@@ -379,20 +453,20 @@ if ($makedef == 1) then
       endif
       echo "read_lef ${lefpath}" >> ${rootname}.cfg
       echo "catch {layers ${route_layers}}" >> ${rootname}.cfg
+      if ( ${?via_use} ) then
+         echo "" >> ${rootname}.cfg
+         echo "via use ${via_use}" >> ${rootname}.cfg
+      endif
       if ( ${?via_pattern} ) then
          echo "" >> ${rootname}.cfg
          echo "via pattern ${via_pattern}" >> ${rootname}.cfg
       endif
       if (! ${?via_stacks} ) then
-         set via_stacks=2
-         echo "via stack ${via_stacks}" >> ${rootname}.cfg
-      endif
-      if ( ${?vddnet} ) then
-	 echo "vdd $vddnet" >> ${rootname}.cfg
-      endif
-      if ( ${?gndnet} ) then
-	 echo "gnd $gndnet" >> ${rootname}.cfg
+         set via_stacks="all"
       endif
+      echo "via stack ${via_stacks}" >> ${rootname}.cfg
+      echo "vdd $vddnet" >> ${rootname}.cfg
+      echo "gnd $gndnet" >> ${rootname}.cfg
 
    else
       echo "# qrouter configuration for project ${rootname}" > ${rootname}.cfg
@@ -406,9 +480,16 @@ if ($makedef == 1) then
          echo "" >> ${rootname}.cfg
          echo "via pattern ${via_pattern}" >> ${rootname}.cfg
       endif
-      if (! ${?via_stacks} ) then
-         set via_stacks=2
-         echo "stack ${via_stacks}" >> ${rootname}.cfg
+      if ( ${?via_stacks} ) then
+         if (${via_stacks} == "none") then
+            echo "no stack" >> ${rootname}.cfg
+	 else
+            if (${via_stacks} == "all") then
+               echo "stack ${route_layers}" >> ${rootname}.cfg
+	    else
+               echo "stack ${via_stacks}" >> ${rootname}.cfg
+	    endif
+	 endif
       endif
    endif
 
@@ -435,9 +516,10 @@ if ($makedef == 1) then
       cat ${rootname}.cfg2 >> ${rootname}.cfg
    else
       if (${scripting} == "T") then
-	 echo "qrouter::standard_route" >> ${rootname}.cfg
-	 # Standard route falls back to the interpreter on failure,
-	 # so make sure that qrouter actually exits.
+	 echo "qrouter::standard_route ${rootname}_route.def false" >> ${rootname}.cfg
+	 echo "qrouter::write_delays ${rootname}.rc" >> ${rootname}.cfg
+	 # Standard_route's automatic quit has been subverted in order
+	 # to write the delay file, so make sure that qrouter actually exits.
 	 echo "quit" >> ${rootname}.cfg
       endif
    endif
@@ -451,8 +533,17 @@ if ($makedef == 1) then
    # netlists.
    #------------------------------------------------------------------
 
+   echo "blifanno.tcl ${synthdir}/${rootname}.blif ${rootname}.def ${synthdir}/${rootname}_anno.blif" \
+	|& tee -a ${synthlog}
    ${scriptdir}/blifanno.tcl ${synthdir}/${rootname}.blif ${rootname}.def \
 		${synthdir}/${rootname}_anno.blif >>& ${synthlog}
+   set errcond = $status
+   if ( ${errcond} != 0 ) then
+      echo "blifanno.tcl failed with exit status ${errcond}" |& tee -a ${synthlog}
+      echo "Premature exit." |& tee -a ${synthlog}
+      echo "Synthesis flow stopped on error condition." >>& ${synthlog}
+      exit 1
+   endif
 
    #------------------------------------------------------------------
    # Spot check:  Did blifanno.tcl produce an output file?
@@ -477,6 +568,16 @@ if ($makedef == 1) then
       echo "" >> ${synthlog}
 
       cd ${synthdir}
+
+      #------------------------------------------------------------------
+      # Copy the original rtl.v and rtlnopwr.v for use in comparison of
+      # pre- and post-placement netlists.
+      #------------------------------------------------------------------
+
+      echo "Copying ${rootname}.rtl.v and ${rootname}.rtlnopwr.v to backups"
+      cp ${rootname}.rtl.v ${rootname}_synth.rtl.v
+      cp ${rootname}.rtlnopwr.v ${rootname}_synth.rtlnopwr.v
+
       echo "Running blif2Verilog." |& tee -a ${synthlog}
       ${bindir}/blif2Verilog -c -v ${vddnet} -g ${gndnet} \
 		${rootname}_anno.blif > ${rootname}.rtl.v
@@ -493,20 +594,20 @@ if ($makedef == 1) then
       # Spot check:  Did blif2Verilog or blif2BSpice exit with an error?
       #------------------------------------------------------------------
 
-      if ( !( -f ${rootname}.rtl.v || \
-		( -M ${rootname}.rtl.v < -M ${rootname}.blif ))) then
+      if ( !( -f ${rootname}.rtl.v || ( -f ${rootname}.rtl.v && \
+		-M ${rootname}.rtl.v < -M ${rootname}.blif ))) then
 	 echo "blif2Verilog failure:  No file ${rootname}.rtl.v created." \
 		|& tee -a ${synthlog}
       endif
 
-      if ( !( -f ${rootname}.rtlnopwr.v || \
-		( -M ${rootname}.rtlnopwr.v < -M ${rootname}.blif ))) then
+      if ( !( -f ${rootname}.rtlnopwr.v || ( -f ${rootname}.rtlnopwr.v && \
+		-M ${rootname}.rtlnopwr.v < -M ${rootname}.blif ))) then
 	 echo "blif2Verilog failure:  No file ${rootname}.rtlnopwr.v created." \
 		|& tee -a ${synthlog}
       endif
 
-      if ( !( -f ${rootname}.spc || \
-		( -M ${rootname}.spc < -M ${rootname}.blif ))) then
+      if ( !( -f ${rootname}.spc || ( -f ${rootname}.spc && \
+		-M ${rootname}.spc < -M ${rootname}.blif ))) then
 	 echo "blif2BSpice failure:  No file ${rootname}.spc created." \
 		|& tee -a ${synthlog}
       endif
