@@ -42,6 +42,11 @@
  *	figured out a way to do this in one stage instead of two, using
  *	the swap_group capability of the graywolf placement tool to find
  *	the optimal groupings.
+ *
+ * Update 6/15/208:
+ *	Finally got around to doing dynamic string allocation on the input,
+ *	which avoids issues of having I/O lines that are arbitrarily long
+ *	crashing the program.
  *---------------------------------------------------------------------------
  *
  * Revision 1.3  2008/09/09 21:24:30  steve_beccue
@@ -66,7 +71,7 @@
 
 #define  FALSE	     0
 #define  TRUE        1
-#define  MAXLINE     2000
+#define  MAXLINE     512
 
 char *Inputfname;
 char *Outputfname;
@@ -238,16 +243,18 @@ int main (int argc, char *argv[])
 {
    int i, j, k;
    int state;
+   int maxline, curline;
    int inputcount;
    int gateinputs;
    int gatecount;
    int doLoadBalance = 1;
    int doFanout = 1;
    char *pinname;
+   char *libfile;
    char *test;
    char *s, *t;
    FILE *infptr, *outfptr;
-   char line[MAXLINE];
+   char *line;
    struct Gatelist *gl = NULL;
    struct Nodelist *nl = NULL, *nlmax, *nlimax;
    struct Drivelist *dl = NULL;
@@ -265,6 +272,8 @@ int main (int argc, char *argv[])
    InitializeHashTable(Drivehash);
    InitializeHashTable(Gatehash);
 
+   fprintf(stdout, "blifFanout for qflow " QFLOW_VERSION "." QFLOW_REVISION "\n");
+
    while ((i = getopt(argc, argv, "fLgnhvl:c:b:i:o:p:s:I:F:")) != EOF) {
       switch (i) {
 	 case 'b':
@@ -277,7 +286,17 @@ int main (int argc, char *argv[])
 	    buf_out_pin = strdup(optarg);
 	    break;
          case 'p':
-	    Gatepath = strdup(optarg);
+	    /* Allow multiple files to be specified as space-separated	*/
+	    /* list, either by specifying all on one "-p" arguments	*/
+	    /* or by passing multiple "-p" arguments.			*/
+	    if (Gatepath == NULL) {
+		Gatepath = strdup(optarg);
+	    }
+	    else {
+		Gatepath = (char *)realloc(Gatepath, strlen(optarg) +
+			strlen(Gatepath) + 2);
+		sprintf(Gatepath, "%s %s", Gatepath, optarg);
+	    }
 	    break;
 	 case 'f':	// fanout only
 	    doLoadBalance = 0;
@@ -311,7 +330,7 @@ int main (int argc, char *argv[])
 	    break;
          case 'h':
 	    helpmessage();
-	    return -3;
+	    return 3;
 	    break;
          default:
 	    break;
@@ -327,7 +346,7 @@ int main (int argc, char *argv[])
       Inputfname = strdup(argv[i]);
       if (!(infptr = fopen(Inputfname, "r"))) {
 	  fprintf(stderr, "blifFanout: Couldn't open %s for reading.\n", Inputfname);
-	  return -1;
+	  return 1;
       }
    }
    i++;
@@ -335,7 +354,7 @@ int main (int argc, char *argv[])
       Outputfname = strdup(argv[i]);
       if (!(outfptr = fopen(Outputfname, "w"))) {
 	 fprintf(stderr, "blifFanout: Couldn't open %s for writing.\n", Outputfname);
-	 return -1;
+	 return 1;
       }
    }
    i++;
@@ -343,9 +362,14 @@ int main (int argc, char *argv[])
    // Make sure we have a valid gate file path
    if (Gatepath == NULL) {
       fprintf(stderr, "blifFanout: No liberty file specified.\n");
-      return -1;
+      return 1;
+   }
+   libfile = strtok(Gatepath, " ");
+   gatecount = 0;
+   while (libfile != NULL) {
+      gatecount += read_gate_file(libfile);
+      libfile = strtok(NULL, " ");
    }
-   gatecount = read_gate_file(Gatepath);
 
    // Make sure we have a non-NULL separator.
    if (Separator == NULL) Separator = &default_sep;
@@ -362,7 +386,7 @@ int main (int argc, char *argv[])
 
    if (gatecount == 0) {
       fprintf(stderr, "blifFanout:  No gates found in %s file!\n", Gatepath);
-      return -1;
+      return 1;
    }
    if (GatePrintFlag) {
       showgatelist();
@@ -426,7 +450,7 @@ int main (int argc, char *argv[])
 	 else
             fprintf(stderr, "blifFanout:  Buffer cell %s cannot be found.\n",
 			Buffername);
-         return -1;
+         return 1;
       }
       for (curpin = gl->gatecell->pins; curpin; curpin = curpin->next) {
 	 if (curpin->type == PIN_INPUT) {
@@ -441,14 +465,25 @@ int main (int argc, char *argv[])
       if (buf_in_pin == NULL || buf_out_pin == NULL) {
 	 fprintf(stderr, "blifFanout:  Could not parse I/O pins "
 			"of buffer cell %s.\n", Buffername);
-	 return -1;
+	 return 1;
       }
    }
 
    pinname = (char *)malloc(1);
    state = NONE;
-   while ((s = fgets(line, MAXLINE, infptr)) != NULL) {
-      t = strtok(s, " \t=\n");
+   maxline = MAXLINE;
+   line = (char *)malloc(MAXLINE * sizeof(char));
+   while ((s = fgets(line, maxline, infptr)) != NULL) {
+      curline = strlen(line);
+      while (curline >= (maxline - 1)) {
+	 maxline += MAXLINE;
+	 line = (char *)realloc(line, maxline * sizeof(char));
+	 s = fgets(line + curline, (maxline - curline), infptr);
+	 curline = strlen(line);
+	 if (s == NULL) break;
+      }
+      if (s == NULL) break;
+      t = strtok(line, " \t=\n");
       while (t) {
 	 switch (state) {
 	    case GATENAME:
@@ -650,10 +685,10 @@ int main (int argc, char *argv[])
    if (infptr != stdin) fclose(infptr);
    if (outfptr != stdout) fclose(outfptr);
 
-   // return with number of gates changed so we can iterate until
-   // this is zero.
+   // Output number of gates changed so we can iterate until this is zero.
 
-   return (Changed_count + Buffer_count);
+   fprintf(stdout, "Number of gates changed: %d\n", Changed_count + Buffer_count);
+   return 0;
 }
 
 /*
@@ -668,7 +703,7 @@ void read_ignore_file(char *ignore_file_name)
 {
    struct Nodelist *nl;
    FILE *ignorefptr;
-   char line[MAXLINE];
+   char line[MAXLINE];	/* One net per line, should not need dynamic allocation */
    char *s, *sp;
 
    if (!(ignorefptr = fopen(ignore_file_name, "r"))) {
@@ -727,7 +762,7 @@ int read_gate_file(char *gate_file_name)
 
 	gl->num_inputs = 0;
 	for (curpin = curcell->pins; curpin; curpin = curpin->next)
-	    if (curpin->type == INPUT)
+	    if (curpin->type == PIN_INPUT)
 		gl->num_inputs++;
 
 	/* The "MaxLatency" is empirically derived.  Since gl->delay	*/
@@ -978,14 +1013,15 @@ void shownodes(void)
 void write_output(int doLoadBalance, FILE *infptr, FILE *outfptr)
 {
    char *s, *t;
-   char line[MAXLINE];
-   char inputline[MAXLINE];
+   char *line;
+   char *inputline;
    char bufferline[MAXLINE];
    char nodename[MAXLINE];
    char *gateline;
    char *stren, *orig;
    int  firstseen;
    int  state, i;
+   int  maxline, curline;
    int  gateinputs;
    int  pincount;
    int  needscorrecting;
@@ -1013,9 +1049,23 @@ void write_output(int doLoadBalance, FILE *infptr, FILE *outfptr)
    gateline = (char *)malloc(1);
    gateline[0] = '\0';
 
-   while ((s = fgets(line, MAXLINE, infptr)) != NULL) {
-      strcpy(inputline, s);		// save this for later
-      t = strtok(s, " \t=\n");
+   maxline = MAXLINE;
+   line = (char *)malloc(MAXLINE * sizeof(char));
+   inputline = (char *)malloc(MAXLINE * sizeof(char));
+
+   while ((s = fgets(line, maxline, infptr)) != NULL) {
+      curline = strlen(line);
+      while (curline >= (maxline - 1)) {
+	 maxline += MAXLINE;
+	 line = (char *)realloc(line, maxline * sizeof(char));
+	 inputline = (char *)realloc(inputline, maxline * sizeof(char));
+	 s = fgets(line + curline, (maxline - curline), infptr);
+	 curline = strlen(line);
+	 if (s == NULL) break;
+      }
+      if (s == NULL) break;
+      strcpy(inputline, line);		// save this for later
+      t = strtok(line, " \t=\n");
       while (t) { 
 	 switch (state) {
 	    case GATENAME:
@@ -1073,12 +1123,17 @@ void write_output(int doLoadBalance, FILE *infptr, FILE *outfptr)
 		  firstseen = 1;
 	       }
 	       gl = (struct Gatelist *)HashLookup(t, Gatehash);
-	       if (gl != NULL) {
-		  gateinputs = gl->num_inputs;
-		  needscorrecting = 0;
-		  pincount = 0;
-		  state = PINNAME;
+	       if (gl == NULL) {
+		  fprintf(stderr, "Fatal error:  Gate \"%s\" is used in source "
+			"but has no liberty file definition.\n", t);
+		  fprintf(stderr, "Check if gate is designated \"dont_use\".\n");
+		  gateinputs = 0;
 	       }
+	       else
+		  gateinputs = gl->num_inputs;
+	       needscorrecting = 0;
+	       pincount = 0;
+	       state = PINNAME;
 	       break;
 	  
 	    case PINNAME:
