@@ -27,15 +27,18 @@ int libCurrentLine;
 char *
 advancetoken(FILE *flib, char delimiter)
 {
-    static char token[LIB_LINE_MAX];
+    static char *token = NULL;
     static char line[LIB_LINE_MAX];
     static char *linepos = NULL;
+    static int token_max_length = LIB_LINE_MAX - 5;
 
     char *lineptr = linepos;
     char *lptr, *tptr;
     char *result;
     int commentblock, concat, nest, quoted;
 
+    if (token == NULL) token = (char *)malloc(LIB_LINE_MAX);
+
     commentblock = 0;
     concat = 0;
     nest = 0;
@@ -53,6 +56,10 @@ advancetoken(FILE *flib, char delimiter)
 	    else lineptr = NULL;
 	}
 
+        // Semicolons are supposed to end lines but sloppy spec allows
+        // them to go missing
+        if (lineptr && *lineptr == '\n' && delimiter == ';') *lineptr = ';';
+
 	if (lineptr == NULL || *lineptr == '\n' || *lineptr == '\0') {
 	    result = fgets(line, LIB_LINE_MAX, flib);
 	    libCurrentLine++;
@@ -66,8 +73,8 @@ advancetoken(FILE *flib, char delimiter)
 		    // backslash and end-of-line, then don't treat as a
 		    // continuation character.
 		    char *eptr = lptr + 1;
-		    while (isspace(*eptr)) eptr++;
-		    if (*eptr == '\0') {
+		    while (isblank(*eptr)) eptr++;
+		    if (*eptr == '\n') {
 			result = fgets(lptr, LIB_LINE_MAX - (lptr - line), flib);
 			libCurrentLine++;
 			if (result == NULL) break;
@@ -84,7 +91,7 @@ advancetoken(FILE *flib, char delimiter)
 
 	if (commentblock == 1) continue;
 
-	while (isspace(*lineptr)) lineptr++;
+	while (isblank(*lineptr)) lineptr++;
 	if (concat == 0)
 	    tptr = token;
 
@@ -110,6 +117,16 @@ advancetoken(FILE *flib, char delimiter)
 		    break;
 	    }
 
+	    // Watch for overruns, and allocate more memory
+	    // for the token if needed.
+
+	    if ((tptr - token) > token_max_length) {
+		char *tsave = token;
+		token_max_length <<= 1;
+		token = (char *)realloc(token, token_max_length);
+		tptr += (token - tsave); 
+	    }
+
 	    // Watch for nested delimiters!
 	    if (delimiter == '}' && *lineptr == '{') nest++;
 	    if (delimiter == ')' && *lineptr == '(') nest++;
@@ -145,12 +162,12 @@ advancetoken(FILE *flib, char delimiter)
     }
     if (delimiter != 0) lineptr++;
 
-    while (isspace(*lineptr)) lineptr++;
+    while (isblank(*lineptr)) lineptr++;
     linepos = lineptr;
 
     // Remove any trailing whitespace
     tptr = token + strlen(token) - 1;
-    while (isspace(*tptr)) {
+    while (isblank(*tptr)) {
 	*tptr = '\0';
 	tptr--;
     }
@@ -337,7 +354,7 @@ get_function(char *out_name, char *lib_func)
 
     while ((sptr = strchr(newfunc, '\'')) != NULL) {
 	fptr = sptr - 1;
-	while (isspace(*fptr)) fptr--;
+	while (isblank(*fptr)) fptr--;
 
 	if (*fptr == ')') {
 	    nest = 1;
@@ -350,7 +367,7 @@ get_function(char *out_name, char *lib_func)
 	}
 	else {
 	    while (*fptr != '!' && *fptr != '*' && *fptr != '+' &&
-		   !isspace(*fptr) && (fptr > newfunc) && *fptr != '('
+		   !isblank(*fptr) && (fptr > newfunc) && *fptr != '('
 		   && *fptr != ')')
 		fptr--;
 	    if (fptr > newfunc) fptr++;
@@ -435,6 +452,7 @@ read_liberty(char *libfile, char *pattern)
     char *libname = NULL;
     int section = INIT;
     LUTable *tables = NULL;
+    BUStype *buses = NULL;
     Cell *cells = NULL;
 
     double time_unit = 1.0;	// Time unit multiplier, to get ps
@@ -445,9 +463,11 @@ read_liberty(char *libfile, char *pattern)
     char *iptr;
 
     LUTable *newtable, *reftable, *scalar;
+    BUStype *newbus, *refbus;
     Cell *newcell, *lastcell;
     Pin *newpin, *lastpin;
     char *curfunc;
+    char *busformat = NULL;
 
     flib = fopen(libfile, "r");
     if (flib == NULL) {
@@ -687,6 +707,7 @@ read_liberty(char *libfile, char *pattern)
 		    if (strcmp(token, "{"))
 			fprintf(stderr, "Error: failed to find start of block\n");
 		    newcell->reftable = NULL;
+		    newcell->bustype = NULL;
 		    newcell->function = NULL;
 		    newcell->pins = NULL;
 		    newcell->area = 1.0;
@@ -749,9 +770,16 @@ read_liberty(char *libfile, char *pattern)
 		   }
 		   cap_unit = strtod(token, &metric);
 		   if (*metric != '\0') {
-		      while (isspace(*metric)) metric++;
+		      while (isblank(*metric)) metric++;
 		      if (*metric == ',') metric++;
-		      while ((*metric != '\0') && isspace(*metric)) metric++;
+		      while ((*metric != '\0') && isblank(*metric)) metric++;
+		      if (*metric == '\"') {
+			 char *qptr;
+			 metric++;
+			 qptr = metric;
+			 while ((*qptr != '\"') && (*qptr != '\0')) qptr++;
+			 *qptr = '\0';
+		      }
 		      if (!strcasecmp(metric, "af"))
 			 cap_unit *= 1E-3;
 		      else if (!strcasecmp(metric, "pf"))
@@ -781,6 +809,53 @@ read_liberty(char *libfile, char *pattern)
 		   }
 		   token = advancetoken(flib, ';');
 		}
+		else if (!strcasecmp(token, "bus_naming_style")) {
+		   token = advancetoken(flib, 0);
+		   if (token == NULL) break;
+		   if (!strcmp(token, ":")) {
+		      token = advancetoken(flib, 0);
+		      if (token == NULL) break;
+		   }
+		   if (!strcmp(token, "\"")) {
+		      token = advancetoken(flib, '\"');
+		      if (token == NULL) break;
+		   }
+		   busformat = strdup(token);
+		   token = advancetoken(flib, ';');
+		}
+		else if (!strcasecmp(token, "type")) {
+		    newbus = (BUStype *)malloc(sizeof(BUStype));
+		    newbus->from = 0;
+		    newbus->to = 0;
+		    newbus->next = buses;
+		    buses = newbus;
+
+		    token = advancetoken(flib, 0);
+		    if (strcmp(token, "("))
+			fprintf(stderr, "Input missing open parens\n");
+		    else
+			token = advancetoken(flib, ')');
+		    newbus->name = strdup(token);
+		    while (*token != '}') {
+			token = advancetoken(flib, 0);
+			if (!strcasecmp(token, "bit_from")) {
+			    token = advancetoken(flib, 0);
+			    token = advancetoken(flib, ';');
+			    sscanf(token, "%d", &newbus->from);
+			}
+			else if (!strcasecmp(token, "bit_to")) {
+			    token = advancetoken(flib, 0);
+			    token = advancetoken(flib, ';');
+			    sscanf(token, "%d", &newbus->to);
+			}
+			else if (!strcmp(token, "{")) {
+			    /* All entries are <name> : <value>.	*/
+			    /* Ignore unhandled tokens.			*/
+			    token = advancetoken(flib, 0);
+			    token = advancetoken(flib, ';');
+			}
+		    }
+		}
 		else {
 		    // For unhandled tokens, read in tokens.  If it is
 		    // a definition or function, read to end-of-line.  If
@@ -805,10 +880,20 @@ read_liberty(char *libfile, char *pattern)
 		    section = LIBBLOCK;			// End of cell def
 		}
 		else if (!strcasecmp(token, "dont_use")) {
-		    free(newcell->name);
-		    newcell->name = NULL;
+		    token = advancetoken(flib, 0);	// Colon
+		    token = advancetoken(flib, ';');	// To end-of-statement
+		    if (!strcasecmp(token, "true")) {
+			char *tmpname;
+			// Prefix name with "**" to mark it as a don't-use cell
+			tmpname = strdup(newcell->name);
+			free(newcell->name);
+			newcell->name = (char *)malloc(strlen(tmpname) + 3);
+			sprintf(newcell->name, "**%s", tmpname);
+			free(tmpname);
+		    }
 		}
-		else if (!strcasecmp(token, "pin")) {
+		else if (!strcasecmp(token, "bus") ||
+				!strcasecmp(token, "pin")) {
 		    token = advancetoken(flib, 0);	// Open parens
 		    if (!strcmp(token, "("))
 			token = advancetoken(flib, ')');	// Close parens
@@ -882,6 +967,20 @@ read_liberty(char *libfile, char *pattern)
 			    fprintf(stderr, "Expected end-of-statement.\n");
 		    }
 		}
+		else if (!strcasecmp(token, "bus_type")) {
+		    token = advancetoken(flib, 0);	// Colon
+		    token = advancetoken(flib, ';');
+		    /* Find the bus */
+		    for (refbus = buses; refbus; refbus = refbus->next)
+			if (!strcmp(refbus->name, token))
+			    break;
+		    if (refbus == NULL)
+			fprintf(stderr, "Failed to find a valid bus type \"%s\"\n",
+				token);
+		    else if (newcell->bustype == NULL) {
+			newcell->bustype = refbus;
+		    }
+		}
 		else if (!strcasecmp(token, "direction")) {
 		    token = advancetoken(flib, 0);	// Colon
 		    token = advancetoken(flib, ';');
@@ -1064,13 +1163,13 @@ read_liberty(char *libfile, char *pattern)
 				    for (i = 0; i < reftable->tsize; i++) {
 					for (j = 0; j < reftable->csize; j++) {
 					    while (*iptr == ' ' || *iptr == '\"' ||
-							*iptr == ',')
+							*iptr == ',' || *iptr == '\\')
 						iptr++;
 					    sscanf(iptr, "%lg", &gval);
 					    *(newcell->values + j * reftable->tsize
 							+ i) = gval * time_unit;
 					    while (*iptr != ' ' && *iptr != '\"' &&
-							*iptr != ',')
+							*iptr != ',' || *iptr == '\\')
 						iptr++;
 					}
 				    }
@@ -1088,21 +1187,28 @@ read_liberty(char *libfile, char *pattern)
 					    *(newcell->values + j * reftable->tsize
 							+ i) = gval * time_unit;
 					    while (*iptr != ' ' && *iptr != '\"' &&
-							*iptr != ',')
+							*iptr != ',' && *iptr != '\0')
 						iptr++;
 					}
 				    }
 				}
 			    }
-
 			    token = advancetoken(flib, 0);
 			    if (strcmp(token, ";"))
 				fprintf(stderr, "Failed to find end of value table\n");
-			    token = advancetoken(flib, 0);
-
 			}
-			else if (strcmp(token, "{"))
-			    fprintf(stderr, "Failed to find end of timing block\n");
+			else if (strcmp(token, "{") && strcmp(token, "}")) {
+			    /* Other tokens:  Unhandled features */
+			    fprintf(stderr, "Unhandled feature %s at line %d\n",
+					token, libCurrentLine);
+			    token = advancetoken(flib, 0);
+			    if (!strcmp(token, "(")) {           // Open parens
+				token = advancetoken(flib, ')'); // Close parens
+				token = advancetoken(flib, 0);
+			    }
+			    if (!strcmp(token, "{"))		 // Open brace
+				token = advancetoken(flib, '}'); // Close brace
+			}
 		    }
 		}
 		else {
@@ -1228,12 +1334,23 @@ int
 get_pintype(Cell *curcell, char *pinname)
 {
     Pin *curpin;
+    char *buschar = NULL;
+
+    /* Is pin part of a bus?  If so, return the root pin */
+    if (curcell->bustype != NULL) {
+	/* Not checking bounds or syntax;  blif requires <...> */
+	buschar = strrchr(pinname, '<');
+	if (buschar != NULL) *buschar = '\0';
+    }
 
     for (curpin = curcell->pins; curpin; curpin = curpin->next) {
 	if (!strcmp(curpin->name, pinname)) {
+	    if (buschar != NULL) *buschar = '<';
 	    return curpin->type;
 	}
     }
+
+    if (buschar != NULL) *buschar = '<';
     return PIN_UNKNOWN;
 }
 
@@ -1243,13 +1360,30 @@ Cell *
 get_cell_by_name(Cell *cell, char *name)
 {
     Cell *currcell, *newcell;
+    char trymore = (char)0;
 
     for (currcell = cell; currcell; currcell = currcell->next) {
         if (!strcasecmp(currcell->name, name)) {
             return currcell;
         }
+	// Quick check on don't-use cells
+	if (currcell->name[0] == '*' && tolower(currcell->name[2]) ==
+			tolower(name[0]))
+	    trymore = (char)1;
     }
-    fprintf(stderr, "Did not find standard cell \"%s\" in list of cells\n", name);
+    if (trymore == (char)1) {
+	for (currcell = cell; currcell; currcell = currcell->next) {
+	    if (*currcell->name == '*' && !strcasecmp(currcell->name + 2, name)) {
+		fprintf(stderr, "Warning: standard cell \"%s\" used but marked "
+			"as dont-use\n", name);
+		// Change cell so that it is no longer marked don't-use
+		strcpy(currcell->name, currcell->name + 2);
+		return currcell;
+	    }
+        }
+    }
+    else
+	fprintf(stderr, "Did not find standard cell \"%s\" in list of cells\n", name);
     return NULL;
 }
 
@@ -1260,7 +1394,7 @@ get_pin_by_name(Cell *curcell, char *pinname)
 
     for (curpin = curcell->pins; curpin; curpin = curpin->next) {
         if (!strcmp(curpin->name, pinname)) {
-            printf("found pin %s\n", pinname);
+            // printf("found pin %s\n", pinname);
             return curpin;
         }
     }
@@ -1269,31 +1403,21 @@ get_pin_by_name(Cell *curcell, char *pinname)
 }
 
 void delete_Cell(Cell *cell) {
-    free(cell->name);
-    free(cell->function);
-    free(cell->times);
-    free(cell->caps);
-    free(cell->values);
-
     Pin *curpin = cell->pins;
     Pin *tmppin;
 
+    if (cell->name != NULL) free(cell->name);
+    if (cell->function != NULL) free(cell->function);
+    if (cell->times != NULL) free(cell->times);
+    if (cell->caps != NULL) free(cell->caps);
+    if (cell->values != NULL) free(cell->values);
+
     while (curpin != NULL) {
         tmppin = curpin->next;
         free(curpin->name);
         free(curpin);
         curpin = tmppin;
     }
-
-    LUTable *curlut = cell->reftable;
-    LUTable *tmplut;
-
-    while (curlut != NULL) {
-        tmplut = curlut->next;
-    /*free(curlut->name);*/
-        /*free(curlut);*/
-        curlut = tmplut;
-    }
     free(cell);
 }
 
@@ -1308,5 +1432,4 @@ delete_cell_list(Cell *cell)
         delete_Cell(currcell);
         currcell = tmpcell;
     }
-
 }
